# coding=iso-8859-15
#
# Newsflash API for external IO.
#
# (c) Sami Visnen 2008
#
# This script provides types and functions for setting
# up a new Newsflash process and "connecting" to it over
# a simple UPD socket. Through this socket commands can be sent
# to newsflash. This effectively allows for simple scripting
# of the host application by sending it sequences of input keys. 
#
# This was originally developed for enable automatic scripted system tests
# thus the documentation is "developer" level. However I belive
# it to be simple enough to allow it to be understood with relative
# easy. 

import socket
import select
import os
import re
import subprocess
from time import sleep
from array import array

# widget and application keys
(VK_MOVE_UP,
VK_MOVE_DOWN,
VK_MOVE_UP_PAGE,
VK_MOVE_DOWN_PAGE,
VK_SET_MARK,
VK_ACTION_SPACE,
VK_ACTION_ENTER,
VK_ERASE,
VK_FOCUS_NEXT,
VK_FOCUS_PREV,
VK_OPEN_MENU,
VK_TAB_COMPLETE_NEXT,
VK_TAB_COMPLETE_PREV,
VK_MOVE_HOME,
VK_MOVE_END,
VK_MOVE_NEXT,
VK_MOVE_PREV,
VK_KILL_LINE,
VK_KILL_CHAR,
VK_KILL_WINDOW,
VK_SENTINEL,
KEY_VIEW_TASKS,
KEY_VIEW_SETTINGS,
KEY_VIEW_HELP,
KEY_VIEW_LOG,
KEY_VIEW_COMMANDS,
KEY_VIEW_SCRIPTS,
KEY_VIEW_STDOUT,
KEY_SERVER_ADD,
KEY_SERVER_KILL,
KEY_SERVER_EDIT,
KEY_SERVER_GET_LIST,
KEY_GROUP_ADD,
KEY_GROUP_KILL,
KEY_GROUP_UPDATE,
KEY_GROUP_CLEAR,
KEY_GROUP_SORT,
KEY_GROUP_PURGE,
KEY_TASK_PAUSE,
KEY_TASK_CANCEL,
KEY_TASK_RESUME,
KEY_TASK_KILL,
KEY_TASK_KILL_TASK,
KEY_TASK_KILL_CONN,
KEY_TASK_SPAWN,
KEY_TASK_FLUSH,
KEY_TASK_VIEW_LOG,
KEY_TASK_VIEW_TASK_LOG,
KEY_TASK_VIEW_CONN_LOG,
KEY_TASK_PRIORITY,
KEY_TASK_YIELD,
KEY_TASK_CLEAR,
KEY_TASK_START,
KEY_CATALOG_FIND,
KEY_CATALOG_SORT,
KEY_CATALOG_REVERSE,
KEY_CATALOG_FILTER,
KEY_CATALOG_BOOKMARK,
KEY_CATALOG_BOOKMARK_NEXT,
KEY_CATALOG_BOOKMARK_PREV,
KEY_CATALOG_JUMP,
KEY_CATALOG_TOUCH,
KEY_CATALOG_INFO,
KEY_EXTENSION_CLOSE,
KEY_EXTENSION_DETAIL,
KEY_EXTENSION_INTERRUPT,
KEY_LIST_SORT_BY_NAME,
KEY_LIST_SORT_BY_SIZE,
KEY_CLEAR_SCREEN,
KEY_SCRIPT_ENABLE,
KEY_EXIT,
) = range(71)

# yes/no/cancel answers
# todo: should fix this somehow so that we didnt have to
# hardwire the actual keys here.
YES    = 'y'
NO     = 'n'
CANCEL = 'c'

# Write output somewhere. By default to stdout
def printlog(msg):
    print msg

class TestError(Exception):
    def __init__(self, msg):
        self.msg = msg

class Screen:
    def __init__(self, width, height, data):
        self.cols = width
        self.rows = height
        self.data = data

    def has_string(self, regex_str):
        regex = re.compile(regex_str)

        for y in range(0, self.rows):
            line=""
            trailing_spaces=0
            for x in reversed(range(0, self.cols)):
                c = self.data[y * self.cols + x]
                if c == 0 or c == 32:
                    trailing_spaces = trailing_spaces + 1
                else:
                    break
            # build the line string
            for x in range(0, self.cols - trailing_spaces):
                c = self.data[y * self.cols + x]
                if c == 0:
                    raise ValueError('unexpected embedded NUL in string')
                line = line + chr(c)

                if re.match(regex, line):
                    return True

        # regex was not found
        return False

class SocketError(Exception):
    def __init__(self, msg):
        self.msg = msg

class Newsflash:
    def __init__(self):
        self.pid = None
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.port = 23555
        self.delay = None
        self.debug = False

    def __del__(self):
        self.s.close()
        self.s = None

    def attach(self, val):
        self.debug = val
        
    def startup(self, binary):
        self.port = self.port + 1
        if os.name == 'nt':
            cmd='cmd /C start ' + binary + ' -port ' + str(self.port) + ' -ignore-session'
            proc=subprocess.Popen(cmd)
            self.pid=proc.pid
        if os.name == 'posix':
            cmd='xterm -e ' + binary + ' -port ' + str(self.port) + ' -ignore-session'
            #os.system(cmd)
            proc=subprocess.Popen(cmd, shell=True)
            self.pid=proc.pid
            
        printlog('started newsflash process:' + str(self.pid))
        # looks like we get a socket error, if trying
        # to send data too fast after just creating the process.
        # thus, hold up a little
        sleep(1)
        if self.debug == True:
            printlog('attach debugger now.... press Enter to continue')
            raw_input()

    def shutdown(self):
        if self.isalive():
            self.sendkey(KEY_EXIT)
            self.sendyes()
            sleep(1)
            
    def kill(self):
        # todo:
        pass

    def command_delay(self, d):
        self.delay = d

    def sendkey(self, key):
        # this helps debugging a testcase
        if self.delay != None:
            sleep(self.delay)
        msg = 'KEY:' + str(key)
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

    def sendraw(self, key):
        msg = 'RAW:' + str(key)[0]
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

    def sendstr(self, str):
        msg = 'STR:' + str
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

    def getsrc(self):
        msg = 'SCREEN'
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

        # assume a screen size up to 100x60.
        # each cell character is represented with a 4byte signed integer.
        # the data string is prefixed with the actual screen resolution.
        data, addr = self.s.recvfrom(100*60*4)
        bytes = len(data)
        #printlog('got %d bytes' % bytes)
        #fmt = '%di' % (bytes / 4)
        #screen = unpack(fmt, data)
        arr = array('i', data)
        width = arr[0]
        height = arr[1]
        arr.pop(0)
        arr.pop(0)
        
        #printlog('width: ' + str(width))
        #printlog('height: ' + str(height))
        return Screen(width, height, arr)

    def sendanswer(self, yesno):
        msg = 'YESNO:' + str(yesno)
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

    # shortcut for host.sendanswer(YES)
    def sendyes(self):
        self.sendanswer(YES)

    def sendno(self):
        self.sendanswer(NO)

    def sendcancel(self):
        self.sendanswer(CANCEL)
    
    def isalive(self):
        if self.pid==None:
            return False
        msg = 'PING'
        if len(msg) != self.s.sendto(msg, ('localhost', self.port)):
            raise SocketError('sendto error')

        read, write, exe = select.select([self.s], [], [], 5)
        if len(read) == 0:
            return False
        # should get PONG back here
        data, addr = self.s.recvfrom(1024)
        return True
